/*
 File:          CSIProcess.h

 Description:   Wrapper for a Chicken interpreter process.

 Author:        Kon Lovett
*/

#import <Foundation/Foundation.h>
#import <OgreKit/OGRegularExpression.h>
#import "CSICommand.h"

@protocol CSIProcessController

// When the result of an action appears on stdout
- (void)csiProcessResultDictionary:(NSDictionary *)outputs forCommand:(CSICommand *)cmd;

// When output appears on stdout
- (void)csiProcessResult:(NSString *)output;

// When output appears on stderr
- (void)csiProcessError:(NSString *)output;

// When csi csiProcess starts
- (void)csiProcessStarted;

// When csi csiProcess quits
- (void)csiProcessFinished;

@end

@interface CSIProcess : NSObject
{
    id <CSIProcessController>   controller;
    NSTask                      *task;
    NSArray                     *arguments;
    NSMutableString             *resultBuffer;
    NSTimer                     *resultTimer;
    NSTimeInterval		resultTimeout;
    NSMutableArray              *commands;
    CSICommand                  *currentCommand;
    OGRegularExpression		*promptRegex;
    int				versionNumber;
    int				buildNumber;
}

// Default launchpath
+ (NSString *)csiDefaultPathname;

// Use default csi launchpath w/o arguments
- (id)initWithController:(id <CSIProcessController>)controller;

// Must specify csi launchpath as the 0th argument
- (id)initWithController:(id <CSIProcessController>)controller arguments:(NSArray *)args;

// Launch arguments, 0 is launchpath
- (NSArray *)arguments;

// The possible values of the 'forCommand:' bit in the
// CSIProcessController csiProcessResultDictionary:forCommand:
// method. Array of CSICommand.
- (NSArray *)commands;

// Launch csi
- (void)startProcess;

// Quit csi
- (void)stopProcess;

// Perform csi action
// result, if any, via csiProcessResultDictionary:forCommand:

- (void)historyAction:(int)num;
- (void)evaluateAction:(NSString *)expr;
- (void)evaluateDescribeAction:(NSArray *)args;
- (void)evaluateDumpAction:(NSArray *)args;
- (void)evaluateDumpRangeAction:(NSArray *)args;
- (void)evaluateTimeAction:(NSArray *)args;
- (void)evaluateMacroPrettyPrintAction:(NSArray *)args;
- (void)loadFilesAction:(NSArray *)args;
- (void)loadFilesVerboseAction:(NSArray *)args;
- (void)reportSystemInfoAction;
- (void)reportTracesAndBreakpointsAction;
- (void)shellCommandAction:(NSArray *)args;
- (void)traceProceduresAction:(NSArray *)args;
- (void)untraceProceduresAction:(NSArray *)args;
- (void)setBreakpointsAction:(NSArray *)args;
- (void)removeBreakpointsAction:(NSArray *)args;
- (void)breakAllThreadsAction;
- (void)breakInThreadAction:(NSArray *)args;
- (void)continueAction;
- (void)describeLastExceptionAction;
- (void)stepAction;
- (void)helpAction;

- (int)versionNumber;
- (int)buildNumber;

// Perform unknown csi action
// result, if any, via csiProcessResult

- (void)performExpression:(NSString *)op arguments:(NSArray *)args;
- (void)performExpression:(NSString *)expr;

@end
