/*
 File:          CSIProcess.m

 Description:   Wrapper for a Chicken interpreter process.

 Author:        Kon Lovett
*/

#import "CSIProcess.h"

/*
    NOTES

	- Assumes Unix newlines.
 
    - Uses a Timer to determine when output from the interpreter process
    is "complete".

    This smells, beyond introducing a fixed latency in results
    csiProcessing.
 
	Timer only used until we know what the csi prompt looks like, so not so bad.
    
    - The actual parsers for the commands are unimplemented.

    - Needs work for modes, like "stepping".
    
    - Could have "unknown" command w/ single "results" entry, instead of
    inline test for nil. Not sure if a good idea though. Current makes
    controller really pay attention to such a situation.
*/

@implementation CSIProcess

enum CommandIndex {
    StartupIndex = 0,
    HistoryIndex,
    EvaluateIndex,
    EvaluatePrettyPrintIndex,
    EvaluateDescribeIndex,
    EvaluateDumpIndex,
    EvaluateDumpRangeIndex,
    EvaluateTimeIndex,
    EvaluateMacroPrettyPrintIndex,
    LoadFilesIndex,
    LoadFilesVerboseIndex,
    ReportSystemInfoIndex,
    ReportTracesAndBreakpointsIndex,
    ShellCommandIndex,
    TraceProceduresIndex,
    UntraceProceduresIndex,
    SetBreakpointsIndex,
    RemoveBreakpointsIndex,
    BreakAllThreadsIndex,
    BreakInThreadIndex,
    ContinueIndex,
    DescribeLastExceptionIndex,
    HelpIndex,
    StepIndex,
    QuitIndex,
    FirstAddedCommandIndex
};

//
// PRIVATE
//

- (BOOL)checkResult:(NSString *)result forCommandIndex:(int)idx rest:(NSRange *)rng
{
	int len;
	CSICommand *cmd = [commands objectAtIndex: idx];
	NSString *op = [cmd name];
	// Must start with the command 'name' echo
	if (![result hasPrefix: op]) {
		NSLog (@"Cannot be result for %@ : %@", cmd, result);
		return NO;
	}
	// Return range of actual result string after echo of 'name\n'
	len = [op length] + 1;
	*rng = NSMakeRange (len, [result length] - len);
	return YES;
}

- (void)oneResultParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [dict setValue: [[result copy] autorelease] forKey: @"result"];
}

- (void)startupParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
	NSArray *lines = [result componentsSeparatedByString: @"\n"];
	int noLines = [lines count];
	int lineNo;
	OGRegularExpression *digits = [OGRegularExpression regularExpressionWithString: @"[0-9]+"];
	OGRegularExpression *specials = [OGRegularExpression regularExpressionWithString: @"([?+*(){}.^$\\\\\\[\\]|])"];
	promptRegex = [[OGRegularExpression regularExpressionWithString: [digits replaceAllMatchesInString: [specials replaceAllMatchesInString: [lines objectAtIndex: noLines - 1] withString: @"\\\\1"] withString: @"[0-9]+"]] retain];
	for (lineNo = 0; lineNo < noLines && ![[lines objectAtIndex: lineNo] hasPrefix: @"Version "]; ++lineNo)
		;
	if (lineNo < noLines) {
		NSScanner *scanner = [NSScanner scannerWithString: [lines objectAtIndex: lineNo]];
		[scanner scanString: @"Version" intoString: nil];
		[scanner scanInt: &versionNumber];
		[scanner scanString: @", Build" intoString: nil];
		[scanner scanInt: &buildNumber];
	}
	[dict setValue: [[[lines subarrayWithRange: NSMakeRange (lineNo + 2, noLines - (lineNo + 2 + 1))]
						componentsJoinedByString: @"\n"] stringByAppendingString: @"\n"]
		  forKey: @"result"];
}

- (void)historyParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluateParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluatePrettyPrintParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self evaluateParse: result intoDictionary: dict];
}

- (void)evaluateDescribeParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluateDumpParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluateDumpRangeParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluateTimeParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)evaluateMacroPrettyPrintParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)loadFilesParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)loadFilesVerboseParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)reportSystemInfoParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
	NSRange rng;
	if (![self checkResult: result forCommandIndex: ReportSystemInfoIndex rest: &rng]) {
		return;
	}
	[dict setValue: [result substringWithRange: rng] forKey: @"result"];
}

- (void)reportTracesAndBreakpointsParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)shellCommandParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)traceProceduresParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)untraceProceduresParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)setBreakpointsParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)removeBreakpointsParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)breakAllThreadsParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)breakInThreadParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)continueParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)describeLastExceptionParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)helpParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)stepParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    [self oneResultParse: result intoDictionary: dict];
}

- (void)quitParse:(NSString *)result intoDictionary:(NSMutableDictionary *)dict
{
    NSLog (@"Error: CSIProcess quitParse:intoDictionary: called");
}

- (CSICommand *)makeCommand:(NSString *)desc name:(NSString *)nam selector:(SEL)sel count:(int)cnt
{
	return [[CSICommand allocWithZone:[self zone]] initWithDescription: desc
                                                   name: nam
                                                   target: self
                                                   selector: sel
                                                   count: cnt];
}

- (id)init
{
    if (self = [super init]) {
		resultTimeout = 0.25;
        resultBuffer = [[NSMutableString allocWithZone:[self zone]] initWithCapacity: 1024]; 
        commands = [[NSMutableArray allocWithZone:[self zone]]
                        initWithObjects:
                            // DO NOT REORDER! (see enum CommandIndex)
                            [self makeCommand: @"startup"
                                  name: nil
                                  selector: @selector(startupParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"history"
                                  name: nil
                                  selector: @selector(historyParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate"
                                  name: nil
                                  selector: @selector(evaluateParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-pretty-print"
                                  name: @",p"
                                  selector: @selector(evaluatePrettyPrintParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-describe"
                                  name: @",d"
                                  selector: @selector(evaluateDescribeParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-dump"
                                  name: @",du"
                                  selector: @selector(evaluateDumpParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-dump-range"
                                  name: @",dur"
                                  selector: @selector(evaluateDumpRangeParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-time"
                                  name: @",t"
                                  selector: @selector(evaluateTimeParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"evaluate-macro-pretty-print"
                                  name: @",x"
                                  selector: @selector(evaluateMacroPrettyPrintParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"load-files"
                                  name: @",l"
                                  selector: @selector(loadFilesParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"load-files-verbose"
                                  name: @",ln"
                                  selector: @selector(loadFilesVerboseParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"report-system-info"
                                  name: @",r"
                                  selector: @selector(reportSystemInfoParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"report-traces-and-breakpoints"
                                  name: @",what"
                                  selector: @selector(reportTracesAndBreakpointsParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"shell-command"
                                  name: @",s"
                                  selector: @selector(shellCommandParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"trace-procedures"
                                  name: @",tr"
                                  selector: @selector(traceProceduresParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"untrace-procedures"
                                  name: @",utr"
                                  selector: @selector(untraceProceduresParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"set-breakpoints"
                                  name: @",br"
                                  selector: @selector(setBreakpointsParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"remove-breakpoints"
                                  name: @",ubr"
                                  selector: @selector(removeBreakpointsParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"break-all-threads"
                                  name: @",breakall"
                                  selector: @selector(breakAllThreadsParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"break-in-thread"
                                  name: @",breakonly"
                                  selector: @selector(breakInThreadParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"continue"
                                  name: @",c"
                                  selector: @selector(continueParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"describe-last-exception"
                                  name: @",exn"
                                  selector: @selector(describeLastExceptionParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"help"
                                  name: @",?"
                                  selector: @selector(helpParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"step"
                                  name: @",step"
                                  selector: @selector(stepParse:intoDictionary:)
                                  count: 1],
                            [self makeCommand: @"quit"
                                  name: @",q"
                                  selector: @selector(quitParse:intoDictionary:)
                                  count: 0],
                            nil];
    }
    return self;
}

- (NSDictionary *)parseResult
{
    return [currentCommand parseResult: resultBuffer];
}

- (void)informController
{
    if (currentCommand) {
        [controller csiProcessResultDictionary: [self parseResult] forCommand: currentCommand];
    } else {
        [controller csiProcessResult: [[resultBuffer copy] autorelease]];
    }
}

- (void)clearResultTimer
{
    // msg to nil ok here but ...
    if (resultTimer) {
        [resultTimer invalidate];
        resultTimer = nil;
    }
}

- (void)resultTimerFired:(NSTimer *)theTimer
{
    [self clearResultTimer];
	[self informController];
}

- (void)clearOutput
{
    [resultBuffer setString: @""];
}

- (void)appendOutput:(NSString *)chunk
{
    [self clearResultTimer];
    [resultBuffer appendString: chunk];
	if (promptRegex) {
		int len = [resultBuffer length], pos;
		NSRange rng;
		// Position at start of last line
		for (pos = len - 1; 0 <= pos && '\n' != [resultBuffer characterAtIndex: pos]; --pos)
			;
		++pos;
		// When prompt found we are done
		rng = NSMakeRange (pos, len - pos);
		if ([promptRegex matchInString: resultBuffer range: rng]) {
			[resultBuffer deleteCharactersInRange: rng];	// Remove the prompt
			[self informController];						// Tell the controller we have a result
		}
	} else {
		resultTimer = [NSTimer scheduledTimerWithTimeInterval: resultTimeout
							   target: self selector: @selector(resultTimerFired:)
							   userInfo: nil
							   repeats: NO];
	}
}

- (void)appendError:(NSString *)chunk
{
    [controller csiProcessError: chunk];
}

- (void)getStdout:(NSNotification *)note
{
    NSData *data = [[note userInfo] objectForKey: NSFileHandleNotificationDataItem];
    if ([data length]) {
        [self appendOutput: [[[NSString allocWithZone:[self zone]] initWithData:data
																   encoding: NSUTF8StringEncoding] autorelease]];
    } else {
        [self stopProcess];
    }
    if (task) {
        [[note object] readInBackgroundAndNotify];
    }
}

- (void)getStderr:(NSNotification *)note
{
    NSData *data = [[note userInfo] objectForKey: NSFileHandleNotificationDataItem];
    if ([data length]) {
        [self appendError: [[[NSString allocWithZone:[self zone]] initWithData:data
																  encoding: NSUTF8StringEncoding] autorelease]];
    } else {
        [self stopProcess];
    }
    if (task) {
        [[note object] readInBackgroundAndNotify];
    }
}

- (id)putStdin:(NSString *)str
{
    if (task && str && 0 != [str length]) {
		NSData *data = [str dataUsingEncoding: NSUTF8StringEncoding allowLossyConversion: NO];
		[[[task standardInput] fileHandleForWriting] writeData: data];
		// data autoreleased?
    }
    return self;
}

- (void)performExpression:(NSString *)expr command:(CSICommand *)cmd
{
    [self clearOutput];
    currentCommand = cmd;
    [[self putStdin: expr] putStdin: @"\n"];
}

- (void)performExpression:(NSString *)op arguments:(NSArray *)args command:(CSICommand *)cmd
{
    NSMutableString *expr = [NSMutableString stringWithCapacity: 20];
    [expr appendString: op];
    if (args) {
        NSEnumerator *enumerator = [args objectEnumerator];
        id arg;
        while (arg = [enumerator nextObject]) {
            [expr appendFormat: @" %@", arg];
        }
    }
    [self performExpression: expr command: cmd];
}

- (void)performExpression:(NSString *)expr commandAtIndex:(unsigned)idx
{
    [self performExpression: expr command: [commands objectAtIndex: idx]];
}

- (void)performCommand:(CSICommand *)cmd arguments:(NSArray *)args
{
    [self performExpression: [cmd name] arguments: args command: cmd];
}

- (void)performCommandAtIndex:(unsigned)idx arguments:(NSArray *)args
{
    [self performCommand: [commands objectAtIndex: idx] arguments: args];
}

- (void)performCommandAtIndex:(unsigned)idx
{
    [self performCommandAtIndex: idx arguments: nil];
}

- (void)quitAction
{
    [self performCommandAtIndex: QuitIndex];
}

//
// PUBLIC
//

+ (NSString *)csiDefaultPathname
{
    return @"/usr/local/bin/csi";
}

- (id)initWithController:(id <CSIProcessController>)cntlr
{
    return [self initWithController: cntlr
                 arguments: [NSArray arrayWithObject: [CSIProcess csiDefaultPathname]]];
}

- (id)initWithController:(id <CSIProcessController>)cntlr arguments:(NSArray *)args
{
    if (self = [self init]) {
        controller = cntlr;
        arguments = [args retain];
    }
    return self;
}

- (void)dealloc
{
    [self stopProcess];
    [commands makeObjectsPerformSelector: @selector(release)];
    [commands release];
    [resultBuffer release];
    [arguments release];
	[promptRegex release];
    [super dealloc];
}

- (NSArray *)arguments
{
    return arguments;
}

- (NSArray *)commands
{
    return commands;
}

- (void)startProcess
{
    task = [[NSTask allocWithZone:[self zone]] init];

    [task setStandardInput: [NSPipe pipe]];
    [task setStandardOutput: [NSPipe pipe]];
    [task setStandardError: [NSPipe pipe]];
    [task setLaunchPath: [arguments objectAtIndex:0]];
    [task setArguments: [arguments subarrayWithRange: NSMakeRange (1, ([arguments count] - 1))]];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                          selector:@selector(getStdout:)
                                          name: NSFileHandleReadCompletionNotification
                                          object: [[task standardOutput] fileHandleForReading]];

    [[[task standardOutput] fileHandleForReading] readInBackgroundAndNotify];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                          selector:@selector(getStderr:)
                                          name: NSFileHandleReadCompletionNotification
                                          object: [[task standardError] fileHandleForReading]];

    [[[task standardError] fileHandleForReading] readInBackgroundAndNotify];

    currentCommand = [commands objectAtIndex: StartupIndex];

    [task launch];

    [controller csiProcessStarted];
}

- (void)stopProcess
{
    if (task) { 
        [controller csiProcessFinished];
        [self quitAction];  
        [self clearResultTimer];    
        [[NSNotificationCenter defaultCenter] removeObserver: self
                                              name: NSFileHandleReadCompletionNotification
                                              object: [[task standardOutput] fileHandleForReading]];    
        [[NSNotificationCenter defaultCenter] removeObserver: self
                                              name: NSFileHandleReadCompletionNotification
                                              object: [[task standardError] fileHandleForReading]]; 
        [task terminate];
        [task release];
        task = nil;
    }
}

- (int)versionNumber
{
	return versionNumber;
}

- (int)buildNumber
{
	return buildNumber;
}

- (NSTimeInterval)actionTimeout
{
	return resultTimeout;
}

- (void)setActionTimeout:(NSTimeInterval)seconds
{
	resultTimeout = seconds;
}

- (void)historyAction:(int)num
{
    [self performExpression: [NSString stringWithFormat: @"#%d", num] commandAtIndex: HistoryIndex];
}

- (void)evaluateAction:(NSString *)expr
{
    [self performExpression: expr commandAtIndex: EvaluateIndex];
}

- (void)evaluateDescribeAction:(NSArray *)args
{
    [self performCommandAtIndex: EvaluateDescribeIndex arguments: args];
}

- (void)evaluateDumpAction:(NSArray *)args
{
    [self performCommandAtIndex: EvaluateDumpIndex arguments: args];
}

- (void)evaluateDumpRangeAction:(NSArray *)args
{
    [self performCommandAtIndex: EvaluateDumpRangeIndex arguments: args];
}

- (void)evaluateTimeAction:(NSArray *)args
{
    [self performCommandAtIndex: EvaluateTimeIndex arguments: args];
}

- (void)evaluateMacroPrettyPrintAction:(NSArray *)args
{
    [self performCommandAtIndex: EvaluateMacroPrettyPrintIndex arguments: args];
}

- (void)loadFilesAction:(NSArray *)args
{
    [self performCommandAtIndex: LoadFilesIndex arguments: args];
}

- (void)loadFilesVerboseAction:(NSArray *)args
{
    [self performCommandAtIndex: LoadFilesVerboseIndex arguments: args];
}

- (void)reportSystemInfoAction
{
    [self performCommandAtIndex: ReportSystemInfoIndex];
}

- (void)reportTracesAndBreakpointsAction
{
    [self performCommandAtIndex: ReportTracesAndBreakpointsIndex];
}

- (void)shellCommandAction:(NSArray *)args
{
    [self performCommandAtIndex: ShellCommandIndex arguments: args];
}

- (void)traceProceduresAction:(NSArray *)args
{
    [self performCommandAtIndex: TraceProceduresIndex arguments: args];
}

- (void)untraceProceduresAction:(NSArray *)args
{
    [self performCommandAtIndex: UntraceProceduresIndex arguments: args];
}

- (void)setBreakpointsAction:(NSArray *)args
{
    [self performCommandAtIndex: SetBreakpointsIndex arguments: args];
}

- (void)removeBreakpointsAction:(NSArray *)args
{
    [self performCommandAtIndex: RemoveBreakpointsIndex arguments: args];
}

- (void)breakAllThreadsAction
{
    [self performCommandAtIndex: BreakAllThreadsIndex];
}

- (void)breakInThreadAction:(NSArray *)args
{
    [self performCommandAtIndex: BreakInThreadIndex arguments: args];
}

- (void)continueAction
{
    [self performCommandAtIndex: ContinueIndex];
}

- (void)describeLastExceptionAction
{
    [self performCommandAtIndex: DescribeLastExceptionIndex];
}

- (void)helpAction
{
    [self performCommandAtIndex: HelpIndex];
}

- (void)stepAction
{
    [self performCommandAtIndex: StepIndex];
}

- (void)performExpression:(NSString *)expr
{
    [self performExpression: expr arguments: nil];
}

- (void)performExpression:(NSString *)op arguments:(NSArray *)args
{
    [self performExpression: op arguments: args command: nil];
}

@end
